#! /usr/bin/env python3
'''Read 4plebs responses and count threads and posts.'''

import sys
import re
from argparse import ArgumentParser
from datetime import datetime, timezone
import json


regexes = {
    'kpg': re.compile('^k-?pop general', re.I),
    'loona': re.compile(r'\bloona\b', re.I),
    'snsd': re.compile(r'\bsnsd\b', re.I),
}

def count_threads(json_paths, year):
    '''Read 4plebs responses and count threads and posts.'''

    seen = set()
    count = {
        k: {
            'threads': 0,
            'posts': 0,
            'images': 0,
            # NOTE that ips only became available midway through 2018, so we'll
            # disregard anything before 2019 when counting them
            'ips': 0 if year >= 2019 else None,
        } for k in list(regexes.keys())+['total']
    }

    cross_year_threads = []
    titles = {k: set() for k in regexes}

    # Jan 1st from the target year until Jan 1st the next year, both at midnight
    # UTC.
    # Note that comparison should be done as start <= timestamp < end
    start = datetime(year, 1, 1, tzinfo=timezone.utc).timestamp()
    end =  datetime(year+1, 1, 1, tzinfo=timezone.utc).timestamp()

    i = 0
    for json_path in json_paths:
        i += 1
        print(f'\r{i}', end='', file=sys.stderr)

        with open(json_path) as f:
            response = json.load(f)

        for thread_num, thread in response.items():
            if thread_num in seen:
                continue
            seen.add(thread_num)

            # check if the thread goes across the new year
            posts = thread.get('posts') or []
            op_time = thread['op']['timestamp']
            op_in_year = start <= op_time < end

            last_post = posts[-1] if posts else thread['op']
            last_post_time = last_post['timestamp']
            last_post_in_year = start <= last_post_time < end

            # disregard thread if fully outside of the target year
            if not (op_in_year or last_post_in_year):
                continue

            if not (op_in_year and last_post_in_year):
                # remember threads that go over the new year
                cross_year_threads.append(thread_num)

                # disregard threads containing posts from the target year but
                # whose OP precedes the new year
                #
                # NOTE that we will be counting posts and images uploaded under
                # a thread from the target year, rather than posts and images
                # uploaded during the target year.
                if not op_in_year:
                    continue

            # calculate post and image counts
            post_count = 1 + thread['omitted'] + len(posts)
            image_count = 1 + thread['images_omitted'] + len(
                [p for p in posts if p['media']])

            to_update = ['total']

            # Check if it's a kpop thread.
            # NOTE that we skip empty duplicate threads here, but we'll still
            # count them towards the board total below
            if post_count > 1:
                title = thread['op']['title'] or thread['op']['comment'] or ''
                for key, regex in regexes.items():
                    # NOTE that up to 2016, there were legitimate KPGs named
                    # "Kpop general" without a '-', but after that, only one
                    # troll thread has that title, which we disregard here
                    if regex.search(title) and not (
                            key == 'kpg' and year > 2016 and
                            title.lower().startswith('kpop')):
                        to_update.append(key)
                        titles[key].add(title)
                        break

            # update counts
            for key in to_update:
                count[key]['threads'] += 1
                count[key]['posts'] += post_count
                count[key]['images'] += image_count
                # disregard ips before 2019 (see note at count declaration
                # above)
                if year >= 2019:
                    count[key]['ips'] += int(thread['op']['unique_ips'])
    print(file=sys.stderr)

    with open(f'{year}_counts.json', 'w') as f:
        json.dump({year: count}, f, indent=2)

    if cross_year_threads:
        with open(f'{year}_cross_year_threads.txt', 'w') as f:
            for thread_num in cross_year_threads:
                print(thread_num, file=f)

    if titles:
        titles = {k: sorted(list(s)) for k, s in titles.items()}
        with open(f'{year}_titles.json', 'w') as f:
            json.dump(titles, f, indent=2)


if __name__ == '__main__':
    _parser = ArgumentParser(description=__doc__)

    _now = datetime.now()

    _parser.add_argument(
        'json', nargs='+',
        help='A 4plebs response encoded as JSON'
    )

    _parser.add_argument(
        '-y', '--year', type=int, default=_now.year-1,
        help='Count threads from this year (default: the previous year)'
    )

    _args = _parser.parse_args()

    count_threads(_args.json, _args.year)
