#! /usr/bin/env python3
'''Generate HTML files from a text files.'''

import sys
import re
from os.path import dirname, realpath, join, exists
import os
from subprocess import run
from html import escape


html_template = '''<!doctype html>
<html>
<head>
<meta charset="UTF-8">
<meta name="robots" content="noindex, nosnippet">
<title>{title}</title>
</head>
<body>
<h1>{title}</h1>
<pre>{body}</pre>
</body>
</html>
'''

# link boundary regex
_bound = r'(?:^|(?<=\s))'

link_regexes = {
    'local_link': {
        'regex': re.compile(
            _bound + r'(\./(?:\w+/)?\w+\.(?:html|conf|lua|json|js|py|sh|7z)\b)',
            re.I
        ),
        'template': (
            r'<a rel="noreferrer noopener" href="\1">\1</a>'
        ),
    },
    'external_link': {
        'regex': re.compile(_bound + r'(https?://\S*[^ (),.*])', re.I),
        'template': (
            r'<a rel="noreferrer noopener" href="\1" target="_blank">\1</a>'
        ),
    },
    'board': {
        'regex': re.compile(_bound + r'(/[a-z]+/)(?=\W)'),
        'template': (
            r'<a rel="noreferrer noopener" href="https://boards.4chan.org\1" '
            r'target="_blank">\1</a>'
        ),
    },
    # must be tested before the 'image' regex below
    'inline_image': {
        'regex': re.compile(
            _bound + r'(\.+/images/inline/(\S*\.(?:jpg|jpeg|png|gif)\b))',
            re.I
        ),
        'template': (
            r'<a href="\1" target="_blank"><img alt="\2" src="\1" '
            r'style="max-width: 800px;"></a>'
        ),
    },
    'image': {
        'regex': re.compile(
            _bound + r'(\.+/images/(\S*\.(?:jpg|jpeg|png|gif)\b))',
            re.I
        ),
        'template': (
            r'<a href="\1" target="_blank"><img alt="\2" src="\1_thumb.jpg"></a>'
        ),
    },
    'strong': {
        'regex': re.compile(r'^[#]([^#]+)[#]$'),
        'template': r'<strong>\1</strong>',
    },
}

board_regex = re.compile(r'\b((/r/)?[a-z]/)\b')

out_dir = join(realpath(dirname(__file__)), 'html')


def generate_html(txt_path):
    '''Generate HTML from a text file.'''
    with open(txt_path) as f:
        lines = [line.strip() for line in f]

    assert lines and len(lines) > 1, f'File is too short: {txt_path}'

    # use the first line as the title
    title = escape(lines.pop(0))

    # drop blank lines after title
    while not lines[0]:
        lines.pop(0)

    # linkify links and escape html entities
    for i, line in enumerate(lines):
        # make sure to escape before replacing links
        line = escape(line)

        for link_type in link_regexes.values():
            line = link_type['regex'].sub(link_type['template'], line)
        lines[i] = line

    # use remaining lines as the body
    body = '\n'.join(lines)

    html = html_template.format(title=title, body=body)

    # generate output file
    components = txt_path.split('/')

    txt_filename = components[-1]
    filename = '{}.html'.format(txt_filename.replace('.txt', ''))

    # allow one subdirectory to be created in the output dir
    subdir = components[-2] if len(components) > 1 else None

    out_subdir = (
        join(out_dir, subdir) if subdir and subdir != 'txt' else out_dir)

    if not exists(out_subdir):
        os.mkdir(out_subdir)

    out_path = join(out_subdir, filename)

    print(f'Generating {out_path}')

    with open(out_path, 'w') as f:
        f.write(html)

    # check output
    p = run(
        ['tidy', '-quiet', '-errors'],
        input=html, capture_output=True, universal_newlines=True, check=False
    )

    if p.returncode > 0:
        print('Errors found in the output html', file=sys.stderr)
        print(p.stderr, file=sys.stderr)


if __name__ == '__main__':
    for _file in sys.argv[1:]:
        generate_html(_file)
