#! /usr/bin/env python3
'''Generate thread statistics for a certain year.'''

import sys
from argparse import ArgumentParser
from datetime import datetime
import json
import locale

import matplotlib.pyplot as plt


locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')


def lfmt(f, v):
    '''Format according to locale.'''
    return locale.format_string(f, v, grouping=True)


def share(y, t):
    '''Board share.'''
    if (y.get(t) and y[t].get('posts') and y.get('total') and
            y['total'].get('posts')):
        return y[t]['posts'] / y['total']['posts'] * 100
    return None


def per_thread(y, t, k, ndigits=None):
    '''Average per thread.'''
    if y.get(t) and y[t].get(k) and y[t].get('threads'):
        return round(y[t][k] / y[t]['threads'], ndigits)
    return None


def posts_per_ip(y, t):
    '''Average posts per IP.'''
    if y.get(t) and y[t].get('ips') and y[t].get('posts'):
        return round(y[t]['posts'] / y[t]['ips'])
    return None


def change(c, p):
    '''Percentage change from previous year.'''
    # c - p is the total difference from previous to current
    # (c - p)/p is the percentage difference
    # = c/p - p/p = c/p - 1
    return (c/p - 1) * 100 if p else None


def print_thread_stats(cur_year, prev_year, tkey, keys, f=sys.stdout):
    '''Calculate and print stats for one thread or for the whole board.'''

    if tkey not in cur_year:
        # nothing to do
        return

    cur = cur_year[tkey]
    prev = prev_year.get(tkey, {})

    for k, label in keys.items():
        # calculate current and previous values
        if k == 'ips':
            c = per_thread(cur_year, tkey, 'ips', ndigits=1)
            p = per_thread(prev_year, tkey, 'ips', ndigits=1)
        else:
            c = cur[k]
            p = prev.get(k)

        if c is None:
            continue

        # format as float or int depending on the key
        fmt = '%.1f' if k in ('share', 'ips') else '%d'
        cf = lfmt(fmt, c) + ('%' if k == 'share' else '')

        # calculate change and per thread value
        _dif = change(c, p)

        pt = (per_thread(cur_year, tkey, k) if k in ('posts', 'images')
              else None)
        pi = posts_per_ip(cur_year, tkey) if k == 'ips' else None

        # output
        print(f'{label}: {cf}', end='', file=f)
        print(f' ({lfmt("%+.1f", _dif)}%)' if _dif else '', end='', file=f)
        # don't print per thread when it's the same as the current count
        # (i.e. if total threads = 1)
        print(f'; {lfmt("%d", pt)} per thread' if pt and pt != c
              else f'; {lfmt("%d", pi)} posts per IP' if pi else '', file=f)


def plot_chart(counts, threads):
    '''Plot chart.'''
    keys = sorted(counts.keys())
    shares = {
        tkey: [
            counts[k].get(tkey, {'share': None})['share']
            for k in keys
        ] for tkey in list(threads) + ['kpop']
    }

    fig, ax = plt.subplots()

    range_str = f'{keys[0]}-{keys[-1]}'
    plt.title(f'/hr/ Board Share ({range_str})')
    plt.xlabel('Year')
    plt.ylabel('Share %')
    plt.rcParams['font.sans-serif'] = 'Arial'
    fig.set_figwidth(10)
    fig.set_figheight(5.625)
    fig.set_layout_engine('tight')

    for tkey, label in threads.items():
        ax.plot(keys, shares[tkey], 'o-', lw=2, label=label)

    ax.fill_between(keys, 0, shares['kpop'], alpha=0.4, label='Total')

    ax.legend(loc='best')

    #plt.show()
    fig.savefig(f'board_share_{range_str}.png', dpi=144)


def generate_stats(json_path, year=None):
    '''Generate thread statistics for a certain year.'''

    if not year:
        # default to last year
        year = datetime.now().year - 1

    threads = {
        'kpg': 'KPG',
        'loona': 'Loona',
        'snsd': 'SNSD',
    }

    with open(json_path) as f:
        counts = json.load(f)

    # calculate board share for each thread and the kpop total
    for y in counts.values():
        y['kpop'] = {'posts': sum(
            (y.get(tkey, {}).get('posts', 0) for tkey in threads))}

        for tkey in list(threads) + ['kpop']:
            if tkey in y:
                y[tkey]['share'] = share(y, tkey)

    cur_year = counts[str(year)]
    prev_year = counts.get(str(year-1), {})

    # print counts
    with open(f'{year}_stats.txt', 'w') as f:
        print(year, end='\n\n', file=f)

        # board
        keys = {
            'posts': 'Total board posts',
            'ips': 'Unique IPs per thread',
        }
        print_thread_stats(cur_year, prev_year, 'total', keys, f=f)

        # individual threads
        keys = {
            'threads': 'Threads',
            'posts': 'Total thread posts',
            'images': 'Total image posts',
            'share': 'Board share',
            'ips': 'Unique IPs per thread',
        }

        for tkey, tlabel in threads.items():
            # skip threads that didn't exist during the target year
            if not cur_year.get(tkey, {'threads': 0}).get('threads'):
                continue

            print(f'\n{tlabel}:', file=f)
            print_thread_stats(cur_year, prev_year, tkey, keys, f=f)

        # total kpop share
        print(file=f)
        keys = {'share': 'Total K-pop board share'}
        print_thread_stats(cur_year, prev_year, 'kpop', keys, f=f)

    # plot chart only if the target year is the last one available (so we don't
    # needlessly redo the same chart over and over when running old stats)
    if str(year) == max(counts.keys()):
        plot_chart(counts, threads)


if __name__ == '__main__':
    _parser = ArgumentParser(description=__doc__)

    _parser.add_argument(
        '--json', default='threads_per_year.json',
        help='Thread counts per year encoded as JSON, '
        '(default: ./threads_per_year.json)'
    )

    _parser.add_argument(
        '-y', '--year', type=int,
        help='The target year (default: the previous year)'
    )

    _args = _parser.parse_args()

    generate_stats(_args.json, year=_args.year)
