#!/usr/bin/env python3
'''Image thumbnailer'''

from os.path import exists, dirname, realpath
import math
from logging import debug

from PIL import Image


_debug = False


class Thumbnailer:
    '''Class for creating square thumbnails centered around the area of highest
    entropy in an image.'''
    # pylint: disable = too-few-public-methods

    @classmethod
    def make_thumb(cls, input_image, output_path, dimension=125):
        '''Save to the output path a square thumbnail of the original image,
        with width and height equal to the specified dimension.

        Return the original image's width and height
        '''

        if not exists(input_image):
            raise FileNotFoundError('File not found: {}'.format(input_image))

        output_path = realpath(output_path)

        if not exists(dirname(output_path)):
            raise FileNotFoundError("Can't save under non-existent directory: "
                                    "{}".format(dirname(output_path)))

        image = Image.open(input_image)

        orig_dimensions = list(image.size)

        # convert anything that is not RGB to RGB (e.g. gifs in palette mode,
        # pngs in RGBA)
        if image.mode != 'RGB':
            image = image.convert('RGB')

        image = cls._resize(image, dimension)

        # crop image using entropy
        cropped_image = cls._crop_to_square_entropy(image)

        cropped_image.save(output_path, quality=90, optimize=True,
                           icc_profile=image.info.get('icc_profile'))

        return orig_dimensions

    @staticmethod
    def _resize(image, dimension):
        '''Resize image so that its smaller dimension will be equal to the
        specified dimension.'''

        width, height = image.size

        if width > height:
            new_height = dimension
            new_width = round(width/height*dimension)
        else:
            new_width = dimension
            new_height = round(height/width*dimension)

        return image.resize((new_width, new_height), resample=Image.LANCZOS)

    @classmethod
    def _crop_to_square_entropy(cls, image):
        '''Crop an image so that it becomes a square. Select the area with the
        greatest entropy in the image.'''

        width, height = image.size
        rotated = False
        if width == height:
            # nothing to be done
            return image
        elif height > width:
            # rotate image so width will be larger than height (will be reversed
            # after cropping)
            rotated = True
            image = image.transpose(Image.ROTATE_90)

        box = cls._get_crop_box_entropy(image)

        if _debug:
            debug('Cropping %sx%s image to %s', width, height, box)

        image = image.crop(box)

        if rotated:
            image = image.transpose(Image.ROTATE_270)
        return image

    @staticmethod
    def _get_centered_area(width, height, target_width=None, target_height=None,
                           center_x=None, center_y=None):
        '''Return rectangle centered around a pixel.'''

        # fill out default values
        target_width = width if target_width is None else target_width
        target_height = height if target_height is None else target_height

        # sanity check
        if target_width > width or target_height > height:
            raise ValueError(
                'Attempt to crop image to a bigger area: {}x{} -> {}x{}'
                ''.format(width, height, target_width, target_height))

        center_x = int(width/2) if center_x is None else center_x
        center_y = int(height/2) if center_y is None else center_y

        # sanity check
        if not (0 <= center_x <= width and 0 <= center_y <= height):
            raise ValueError(
                'Attempt to center outside of the image area: '
                '{}x{} not in {}x{}'.format(center_x, center_y, width, height))

        if _debug:
            debug('Centering %sx%s -> %sx%s around (%s, %s)', width, height,
                  target_width, target_height, center_x, center_y)

        # center horizontally
        left = (center_x - int(target_width/2) if target_width < width else 0)
        right = (left + target_width if target_width < width else width)

        # center vertically
        top = (center_y - int(target_height/2) if target_height < height else 0)
        bottom = (top + target_height if target_height < height else height)

        # if any values are beyond the image's boundaries, move them back in
        if left < 0:
            right -= left
            left = 0
        elif right > width:
            left -= right-width
            right = width

        if top < 0:
            bottom -= top
            top = 0
        elif bottom > height:
            top -= bottom-height
            bottom = height

        return (left, top, right, bottom)

    @classmethod
    def _get_crop_box_entropy(cls, image):
        '''Select best square area for the image.

        The image's width must be greater than the image's height.
        '''

        width, height = image.size
        # the target width is equal to the height
        target = height

        width_center = cls._get_center_of_entropy(image)

        if _debug:
            debug('%s %s', width_center,
                  '{}%'.format(int(width_center/width*100)))

        return cls._get_centered_area(width, height, target_width=target,
                                      center_x=width_center)

    @classmethod
    def _get_center_of_entropy(cls, image):
        '''Calculates the center of entropy (the point at which there's as much
        entropy on either side) for the image's width.'''
        width, height = image.size

        entropies = []

        for i in range(width):
            region = image.crop((i, 0, i+1, height))
            entropy = cls._calculate_entropy(region)
            entropies.append(entropy)

        #if _debug:
        #    debug(entropies)

        return cls._get_array_center(entropies)

    @staticmethod
    def _calculate_entropy(region):
        '''Calculates the entropy for a 1-pixel wide strip from an image.

        The formula used for entropy is:

        entropy = - sum(p(i)*log2(p(i)))

        where sum() is the sum for all values of i; i is any possible
        difference between two adjacent pixels; and p(i) is the probability of a
        difference being equal to i.
        '''
        occurrences = {}
        height = region.size[1]

        # calculate and store the number of occurrences for the difference in
        # values between any two adjacent pixels
        for i in range(height-1):
            p1 = region.getpixel((0, i))
            p2 = region.getpixel((0, i+1))

            if isinstance(p1, int):
                # greyscale
                diff = abs(p2-p1)
            else:
                # RGB
                # difference is the sum of differences for each band
                diff = abs(p2[0]-p1[0]) + abs(p2[1]-p1[1]) + abs(p2[2]-p1[2])

            if not diff in occurrences:
                occurrences[diff] = 1
            else:
                occurrences[diff] += 1

        # calculate entropy
        entropy = 0
        for diff in occurrences:
            # probability = occurrences / total occurrences
            probability = occurrences[diff]/(height-1)

            entropy -= probability*math.log2(probability)

        return entropy

    @staticmethod
    def _get_array_center(array):
        '''Returns the index for the center of mass of an array.

        The formula used for the center of mass is

        center = 1/M*sum(m(i)*i)

        where sum is the sum for all values of i; i is each position in the
        array; M is the total mass of the array; and m(i) is the value of the
        mass located at i.
        '''

        total_mass = 0
        for value in array:
            total_mass += value

        # calculate the weighted sum for all positions
        weighted_sum = 0
        for i, value in enumerate(array):
            weighted_sum += i*value

        # avoid division by zero
        if total_mass*weighted_sum != 0:
            center = 1/total_mass*weighted_sum
        else:
            # this should only happen for images with zero entropy, so the
            # center doesn't actually matter since anything will produce the
            # exact same thumbnail
            center = 0

        if _debug:
            debug(center)

        return round(center)


if __name__ == '__main__':
    import sys

    for _file in sys.argv[1:]:
        _thumb = f'{_file}_thumb.jpg'
        Thumbnailer.make_thumb(_file, _thumb)
